"""
Import all possible stuff that can be used.
"""


from ._lib_info import (
    have_decoder_for_format,
    have_encoder_for_format,
    libheif_info,
    libheif_version,
)
from ._options import options
from ._version import __version__
from .as_opener import HeifImageFile, register_heif_opener
from .constants import heif_filetype_maybe  # DEPRECATED
from .constants import heif_filetype_no  # DEPRECATED
from .constants import heif_filetype_yes_supported  # DEPRECATED
from .constants import heif_filetype_yes_unsupported  # DEPRECATED
from .constants import (
    HeifChannel,
    HeifChroma,
    HeifColorProfileType,
    HeifColorspace,
    HeifCompressionFormat,
    HeifErrorCode,
    HeifFiletype,
)
from .error import HeifError

# pylint: disable=redefined-builtin
from .heif import (
    HeifFile,
    HeifImage,
    HeifThumbnail,
    check,
    check_heif,
    from_bytes,
    from_pillow,
    is_supported,
    open,
    open_heif,
    read,
    read_heif,
)
from .misc import get_file_mimetype, getxmp, set_orientation
from .thumbnails import add_thumbnails, thumbnail
